from enum import Enum


class EventType(Enum):
    """Enum of currently supported event_types"""

    DATA = "data"
    STATE = "state"
    CONFIG = "config"
    CMD = "cmd"
    HEARTBEAT = "heartbeat"



import copy
from ebb_events.event_schema import DataSchema
from ebb_events.field_constants import METADATA


def validate_data_event_payload_message(payload_message: dict):
    """
    Helper returns True if data payload_message matches expected structure.
    Helper returns dictionary of error information indicating which fields in the
    payload_message do not follow the expected structure.

    Expected structure:
    {
        "var_name": {
            "value": float,
            "units": str,
        },
        ...
    }
    """
    # Copy of payload_message so that we don't actually alter the raw_payload itself
    payload_message_copy: dict = copy.deepcopy(payload_message)
    payload_message_copy.pop(METADATA, None)
    errors_dict = DataSchema(many=True).validate(list(payload_message_copy.values()))
    return True if errors_dict == {} else errors_dict



import copy
import logging
from datetime import datetime
from marshmallow import ValidationError
from ebb_events.field_constants import (
    DATA,
    ID,
    METADATA,
    SERIAL_NUMBER,
    SOURCE,
    TIME,
    TYPE,
)
from ebb_events.builders.event_builder import EventEnvelope
from ebb_events.event_schema import EventPayloadSchema
from ebb_events.exceptions import PayloadFormatException


class EventConsumer:
    """
    Class to consume and parse events that follow the ebb_event structure
    """

    def __init__(self, payload: dict = {}) -> None:
        """
        Initializes an EventConsumer object with payload data
        """
        self.error_log_message = (
            "Error fetching %s. Invalid payload, not an ebb-event structure."
        )
        self.raw_payload: dict = payload
        self.is_ebb_event_structure: bool = self.check_is_ebb_event_structure(
            payload=self.raw_payload
        )

    def check_is_ebb_event_structure(self, payload: dict) -> bool:
        """
        Helper checks the payload structure to see if it matches our EventPayloadSchema
        """
        payload_schema = EventPayloadSchema()
        try:
            payload_schema.load(data=payload)
            return True
        except ValidationError:
            return False

    def get_event_id(self) -> str:
        """
        Helper to fetch event's id field as str(id).

        Raises:
            PayloadFormatException: Raised if event does not match ebb_event structure.

        Returns:
            str: event's string id.
        """
        if self.is_ebb_event_structure:
            return str(self.raw_payload.get(ID))
        logging.error(self.error_log_message % "event_id")
        raise PayloadFormatException()

    def get_event_time_str(self) -> str:
        """
        Helper to fetch event's time field in RFC3339 format: YYYY-MM-DDThh:mm:ss.ssssss+/-hh:mm

        Raises:
            PayloadFormatException: Raised if event does not match ebb_event structure.

        Returns:
            str: event's time field.
        """
        if self.is_ebb_event_structure:
            return str(self.raw_payload.get(TIME))
        logging.error(self.error_log_message % "event_time_str")
        raise PayloadFormatException()

    def get_event_time(self) -> datetime:
        """
        Helper to fetch event's time field as python datetime object.

        Raises:
            PayloadFormatException: Raised if event does not match ebb_event structure.

        Returns:
            datetime: event's time field.
        """
        if self.is_ebb_event_structure:
            raw_time = self.raw_payload.get(TIME)
            try:
                return datetime.fromisoformat(raw_time)
            except ValueError as error:
                logging.error(
                    f"Time is invalid isoformat string: {raw_time}",
                    extra={"error": str(error)},
                )
                raise PayloadFormatException(
                    message=f"Time is invalid isoformat string: {raw_time}"
                ) from error
        logging.error(self.error_log_message % "event_time")
        raise PayloadFormatException()

    def get_event_source(self) -> str:
        """
        Helper to fetch event's source as string. For ebb_events, the `source` field matches the MQTT `topic`.

        Raises:
            PayloadFormatException: Raised if event does not match ebb_event structure.

        Returns:
            str: event's source field.
        """
        if self.is_ebb_event_structure:
            return str(self.raw_payload.get(SOURCE))
        logging.error(self.error_log_message % "event_source/event_topic")
        raise PayloadFormatException()

    def get_event_topic(self) -> str:
        """
        Helper to fetch event's topic string. For ebb_events, the MQTT `topic` field matches the event's source.

        Raises:
            PayloadFormatException: Raised if event does not match ebb_event structure.

        Returns:
            str: event's topic string.
        """
        return self.get_event_source()

    def get_event_type(self) -> str:
        """
        Helper to fetch event's type as string.

        Raises:
            PayloadFormatException: Raised if event does not match ebb_event structure.

        Returns:
            str: event's type field.
        """
        if self.is_ebb_event_structure:
            return str(self.raw_payload.get(TYPE))
        logging.error(self.error_log_message % "event_type")
        raise PayloadFormatException()

    def get_event_organization(self) -> str:
        """
        Helper to fetch event's organization portion of topic hierarchy as string.
        This is the top level of the topic hierarchy.

        Raises:
            PayloadFormatException: Raised if event does not match ebb_event structure.

        Returns:
            str: event's organization portion of the topic hierarchy.
        """
        # This ensures that the `source` field matches the expected topic structure
        if self.is_ebb_event_structure:
            topic_list = self.raw_payload.get(SOURCE).split("/")
            return topic_list[0]
        logging.error(self.error_log_message % "event_organization")
        raise PayloadFormatException()

    def get_event_system_id(self) -> str:
        """
        Helper to fetch event's system_id portion of topic hierarchy as string.
        This is the second level of the topic hierarchy.

        Raises:
            PayloadFormatException: Raised if event does not match ebb_event structure.

        Returns:
            str: event's system_id portion of the topic hierarchy.
        """
        # This ensures that the `source` field matches the expected topic structure
        if self.is_ebb_event_structure:
            topic_list = self.raw_payload.get(SOURCE).split("/")
            return topic_list[1]
        logging.error(self.error_log_message % "event_system_id")
        raise PayloadFormatException()

    def get_event_subsystem_id(self) -> str:
        """
        Helper to fetch event's subsystem_id portion of topic hierarchy as string.
        This is the fourth level of the topic hierarchy.

        Raises:
            PayloadFormatException: Raised if event does not match ebb_event structure.

        Returns:
            str: event's subsystem_id portion of the topic hierarchy.
        """
        # This ensures that the `source` field matches the expected topic structure
        if self.is_ebb_event_structure:
            topic_list = self.raw_payload.get(SOURCE).split("/")
            return topic_list[3]
        logging.error(self.error_log_message % "event_subsystem_id")
        raise PayloadFormatException()

    def get_event_device_id(self) -> str:
        """
        Helper to fetch event's device_id portion of topic hierarchy as string.
        This is the fourth level of the topic hierarchy.

        Raises:
            PayloadFormatException: Raised if event does not match ebb_event structure.

        Returns:
            str: event's device_id portion of the topic hierarchy.
        """
        # This ensures that the `source` field matches the expected topic structure
        if self.is_ebb_event_structure:
            topic_list = self.raw_payload.get(SOURCE).split("/")
            return topic_list[4]
        logging.error(self.error_log_message % "event_device_id")
        raise PayloadFormatException()

    def get_event_envelope(self) -> EventEnvelope:
        """
        Helper to build and return EventEnvelope from payload.

        Raises:
            PayloadFormatException: Raised if event does not match ebb_event structure.

        Returns:
            EventEnvelope: event envelope object with expected fields.
        """
        if self.is_ebb_event_structure:
            return EventEnvelope(
                organization=self.get_event_organization(),
                system_id=self.get_event_system_id(),
                event_type=self.get_event_type(),
                subsystem_id=self.get_event_subsystem_id(),
                device_id=self.get_event_device_id(),
            )
        logging.error(self.error_log_message % "event_envelope")
        raise PayloadFormatException()

    def get_event_message(self, metadata_included=True) -> dict:
        """
        Helper fetches and returns message dict from payload.
        Metadata field in the message dict is included by default but can be parsed out.

        Args:
            metadata_included (bool, optional): If set to False, returned dict will remove the `metadata` portion of the message.
                                                Defaults to True.

        Raises:
            PayloadFormatException: Raised if event does not match ebb_event structure.

        Returns:
            dict: Message value of payload with or without the `metadata` field depending on metadata_included parameter.
        """
        if self.is_ebb_event_structure:
            # Copy of message so that we don't actually alter the raw_payload itself
            message: dict = copy.deepcopy(self.raw_payload.get(DATA))
            if not metadata_included:
                message.pop(METADATA, {})
            return message
        logging.error(self.error_log_message % "event_message")
        raise PayloadFormatException()

    def get_event_message_metadata(self) -> dict:
        """
        Helper to fetch and return the metadata dictionary within the payload's event message.

        Raises:
            PayloadFormatException: Raised if event does not match ebb_event structure.

        Returns:
            dict: Metadata included in event message.
        """
        if self.is_ebb_event_structure:
            return self.raw_payload.get(DATA, {}).get(METADATA, {})
        logging.error(self.error_log_message % "event_message_metadata")
        raise PayloadFormatException()

    def get_device_serial_number(self) -> str:
        """
        Helper to fetch and return the serial number of the publishing device as a string.
        Pulls this information from the metadata of the event message.

        Raises:
            PayloadFormatException: Raised if event does not match ebb_event structure.

        Returns:
            str: Serial number from metadata of event if present, otherwise None
        """
        if self.is_ebb_event_structure:
            if (
                ser_num := self.raw_payload.get(DATA, {})
                .get(METADATA, {})
                .get(SERIAL_NUMBER, None)
            ):
                return str(ser_num)
            return None
        logging.error(self.error_log_message % "device_serial_number")
        raise PayloadFormatException()



class TopicFormatException(Exception):
    """Exception type raised for invalid topic structure/format"""

    pass


class PayloadFormatException(Exception):
    """Exception type raised for invalid payload structure/format"""

    def __init__(self, message="Payload does not match required ebb_event format"):
        self.message = message
        super().__init__(self.message)



TOPIC_HIERARCHY_FIELDS = [
    "organization",
    "system_id",
    "event_type",
    "subsystem_id",
    "device_id",
]

# Payload fields
ID = "id"
TIME = "time"
SOURCE = "source"
TYPE = "type"
DATA = "data"
METADATA = "metadata"
SERIAL_NUMBER = "serial_number"



from marshmallow import Schema, fields, validate

from ebb_events.enums import EventType


TOPIC_MAX_LENGTH = 50
TOPIC_REGEX = r"^[a-z0-9-]+$"
SOURCE_REGEX = r"^([a-z0-9-]+\/){4}[a-z0-9-]+$"
SOURCE_MAX_LENGTH = 256


def remove_nones_helper(data):
    """
    Recursive helper returns the parameter dictionary with None key:values removed
    from top level and all nested dictionaries
    """
    if isinstance(data, dict):
        return {
            key: remove_nones_helper(value)
            for key, value in data.items()
            if value is not None and remove_nones_helper(value) is not None
        }
    return data


class EventEnvelopeSchema(Schema):
    """Schema for EventEnvelope class to validate fields follow requirements"""

    organization = fields.Str(
        validate=[
            validate.Length(max=TOPIC_MAX_LENGTH),
            validate.Regexp(regex=TOPIC_REGEX),
        ]
    )
    system_id = fields.Str(
        validate=[
            validate.Length(max=TOPIC_MAX_LENGTH),
            validate.Regexp(regex=TOPIC_REGEX),
        ]
    )
    event_type = fields.Str(
        validate=[
            validate.Length(max=TOPIC_MAX_LENGTH),
            validate.OneOf([e.value for e in EventType]),
        ]
    )
    subsystem_id = fields.Str(
        validate=[
            validate.Length(max=TOPIC_MAX_LENGTH),
            validate.Regexp(regex=TOPIC_REGEX),
        ]
    )
    device_id = fields.Str(
        validate=[
            validate.Length(max=TOPIC_MAX_LENGTH),
            validate.Regexp(regex=TOPIC_REGEX),
        ]
    )


class RawToStringField(fields.Raw):
    """
    Class for our EventPayloadSchema id field to accept any id but returns
    that id as a string value e.g. (str(id))
    """

    def _serialize(self, value, attr, obj, **kwargs):
        return str(value)


class DictConditionallyRemoveNone(fields.Dict):
    """
    Class for EventPayloadSchema data field to conditionally
    remove None fields
    """

    def _serialize(self, value, attr, obj, **kwargs):
        remove_nones = self.context.get("remove_nones", False)
        if remove_nones:
            # Filter out None values from the dictionary for top and second levels
            filtered_dict = remove_nones_helper(value)
            return filtered_dict
        else:
            return value


class EventPayloadSchema(Schema):
    """Schema for event payload json to validate fields follow requirements"""

    # id field could be uuid but doesn't have to be (could be string(int))
    id = RawToStringField(required=True)
    time = fields.DateTime(required=True, format="iso")
    source = fields.Str(
        required=True,
        validate=[
            validate.Length(max=SOURCE_MAX_LENGTH),
            validate.Regexp(regex=SOURCE_REGEX),
        ],
    )
    type = fields.Str(
        required=True,
        validate=[
            validate.OneOf([e.value for e in EventType]),
        ],
    )
    data = DictConditionallyRemoveNone(required=True)


class DataSchema(Schema):
    """Schema for event data entries to check expected structure"""

    value = fields.Float(required=True)
    units = fields.Str(required=False, allow_none=True)



import json
import logging

from datetime import datetime, timezone
from marshmallow import ValidationError
from typing import Optional
from uuid import uuid4

from ebb_events.enums import EventType
from ebb_events.field_constants import (
    DATA,
    ID,
    METADATA,
    SOURCE,
    TIME,
    TYPE,
    SERIAL_NUMBER,
)
from ebb_events.event_schema import EventEnvelopeSchema, EventPayloadSchema
from ebb_events.exceptions import PayloadFormatException, TopicFormatException


class EventEnvelope:
    """
    EventEnvelope class used to build the event payload and topic
    """

    def __init__(
        self,
        organization: str,
        system_id: str,
        event_type: str,
        subsystem_id: str,
        device_id: str,
    ) -> None:
        """Initialize an EventEnvelope class object with required fields and validate those fields"""
        self.organization = organization
        self.system_id = system_id
        self.event_type = event_type
        self.subsystem_id = subsystem_id
        self.device_id = device_id
        # Fields must all be valid to initialize this class
        self._validate()

    def __eq__(self, other):
        """Write the __eq__ method to evaluate EventEnvelopes with equivalent fields"""
        if not isinstance(other, EventEnvelope):
            return False
        return (
            self.organization == other.organization
            and self.system_id == other.system_id
            and self.event_type == other.event_type
            and self.subsystem_id == other.subsystem_id
            and self.device_id == other.device_id
        )

    def _validate(self):
        """
        Validate that the fields passed to EventEnvelope follow expected format and rules.

        Return: None if valid fields
        Raises: TopicFormatException for invalid fields
        """
        envelope_schema = EventEnvelopeSchema()
        errors = envelope_schema.validate(envelope_schema.dump(self))
        if errors:
            # errors will be a dictionary of validation errors
            logging.error("Invalid EventEnvelope", extra={"errors": errors})
            raise TopicFormatException(errors)

    def _format_time(self, datetime_raw: Optional[datetime] = None) -> str:
        """
        Helper takes in datetime object or None and returns the string datetime in RFC3339 format UTC.
        If None - timestamp is datetime.now().

        Format: YYYY-MM-DDThh:mm:ss.ssssss+/-hh:mm
        """
        datetime_raw = (
            datetime_raw if datetime_raw is not None else datetime.now(timezone.utc)
        )
        if (
            datetime_raw.tzinfo is None
            or datetime_raw.tzinfo.utcoffset(datetime_raw) is None
        ):
            datetime_raw = datetime_raw.replace(tzinfo=timezone.utc)
        return datetime_raw.isoformat()

    def build_event_topic(self, is_heartbeat: bool = False) -> str:
        """
        Joins the EventEnvelope fields to form a valid topic string.
        Note: Fields are validated in __init__ so no need to re-validate here.

        args:
            is_heartbeat (bool): If True, replace event_type section of topic with EventType.HEARTBEAT.
                                Defaults to False.
        Returns:
            str: _description_
        """
        topic = f"{self.organization}/{self.system_id}/{self.event_type if is_heartbeat is False else EventType.HEARTBEAT.value}/{self.subsystem_id}/{self.device_id}"
        # This should be enforced by the ._validate() call and the EventEnvelopeSchema too
        if len(topic) > 256:
            logging.error(f"Invalid Topic. Topic too long: {topic}.")
            raise TopicFormatException("Topic length cannot exceed 256 characters.")
        return topic

    def build_event_payload_dict(
        self,
        message: dict,
        serial_number: str = None,
        metadata: dict = {},
        datetime_obj: Optional[datetime] = None,
        remove_nones: bool = False,
    ) -> dict:
        """
        Method validates the message structure, builds and returns the expected
        event payload to be used in the MQTT payload as a json serializable dictionary.
        NOTE: Returned object is a python dictionary, not a json object. In order to publish this,
        one must either `json.dumps()` this result or use `build_event_payload_json()` method.

        Args:
            message (dict): Dictionary containing the message information that is being published.
            serial_number (str): String serial number of the sensor publishing this event
                    This will be added to metadata dict and will overwrite any metadata field
                    named "serial_number".
            metadata (dict): Dictionary containing extra metadata information to be included in the data.
            datetime_obj (Optional[datetime]): datetime obj to use as event time.
            remove_nones (bool): If set to True, all key:value pairs where value=None will
                                    be removed/omitted from the resulting payload dict

        Returns:
            dict: Dictionary of the expected MQTT event message format.

        Exceptions:
            Raises 'PayloadFormatException' if the topic or message does not meet the expected format.
        """
        if serial_number is not None:
            metadata[SERIAL_NUMBER] = str(serial_number)
        payload_schema = EventPayloadSchema(context={"remove_nones": remove_nones})
        if datetime_obj is not None and type(datetime_obj) is not datetime:
            logging.error(
                f"Invalid datetime_obj: {datetime_obj} must be of type datetime."
            )
            raise PayloadFormatException(
                "datetime_obj must be a valid datetime object."
            )
        try:
            event_payload = {
                ID: uuid4(),
                TIME: self._format_time(datetime_raw=datetime_obj),
                SOURCE: self.build_event_topic(),
                TYPE: self.event_type,
                DATA: {METADATA: metadata, **message},
            }
            # validate that data and metadata are JSON serializable
            json.dumps(message)
            json.dumps(metadata)
            deserialized_payload = payload_schema.load(
                data=event_payload
            )  # validates event_payload against schema
        except TopicFormatException:
            # Exception raised by build_event_topic() if too long
            # logging.error called at source of the TopicFormatException raised
            raise PayloadFormatException(
                "Source and topic must be less than 256 characters."
            )
        except ValidationError as error:
            # Exception raised by schema.load() if not valid marshmallow schema
            logging.error(
                f"Invalid payload. Does not match marshmallow schema: {str(error)}"
            )
            raise PayloadFormatException(
                f"Payload does not match required format: {str(error)}"
            )
        except TypeError as error:
            # Exception raised by json.dumps() if not valid json serializable object
            logging.error(f"Invalid payload, not JSON serializable: {str(error)}")
            raise PayloadFormatException(
                "Payload message and metadata must be JSON serializable dictionaries."
            )
        return payload_schema.dump(deserialized_payload)

    def build_event_payload_json(
        self,
        message: dict,
        serial_number: str = None,
        metadata: dict = {},
        datetime_obj: Optional[datetime] = None,
        remove_nones: bool = False,
    ):
        """
        Method validates the message structure, builds and returns the expected
        event payload to be used in the MQTT payload as a JSON formatted str.

        This returned value can be immediately placed in an MQTT publish call as it's a json string.

        Args:
            message (dict): Dictionary containing the message information that is being published.
            serial_number (str): String serial number of the sensor publishing this event
                    This will be added to metadata dict and will overwrite any metadata field
                    named "serial_number".
            metadata (dict): Dictionary containing extra metadata information to be included in the data
            datetime_obj (Optional[datetime]): datetime obj to use as event time.
            remove_nones (bool): If set to True, all key:value pairs where value=None will
                                    be removed/omitted from the resulting payload json.

        Returns:
            dict: Dictionary of the expected MQTT event message format.

        Exceptions:
            Raises 'PayloadFormatException' if the topic or message does not meet the expected format.
        """
        payload_dict = self.build_event_payload_dict(
            message=message,
            serial_number=serial_number,
            metadata=metadata,
            datetime_obj=datetime_obj,
            remove_nones=remove_nones,
        )
        return json.dumps(payload_dict)

    def build_event_heartbeat_payload_json(
        self,
        metadata: dict = {},
    ) -> str:
        """
        Build a heartbeat event payload to publish

        Args:
            metadata (dict, optional): Dictionary containing extra metadata information to be included in the data. Defaults to {}.

        Returns:
            str: json payload for heartbeat messages to publish
        """
        # Build payload with {"alive": True} heartbeat data message
        payload_dict = self.build_event_payload_dict(
            message={"alive": True}, metadata=metadata
        )

        # Overwrite SOURCE and TYPE with heartbeat specific values
        heartbeat_topic = self.build_event_topic(is_heartbeat=True)
        payload_dict[SOURCE] = heartbeat_topic
        payload_dict[TYPE] = EventType.HEARTBEAT.value
        return json.dumps(payload_dict)
